package main

import (
	"fmt"
	"os"
	"time"

	"github.com/jessevdk/go-flags"

	"github.com/go-git/go-git/v5"
)

var arguments *commandLineOptions
var gitRepo *git.Repository

type commandLineOptions struct {
	Path     string `short:"p" long:"path" description:"path to the file"`
	Selector string `short:"s" long:"selector" description:"selector of data"`
	Git      bool   `short:"g" long:"git" description:"read from git"`
	Verbose  bool   `short:"v" long:"verbose" description:"verbose output"`
	Null     bool   `short:"0" long:"null" description:"terminate with null byte"`

	Major struct {
	} `command:"major" description:"increase the major version"`
	Minor struct {
	} `command:"minor" description:"increase the minor version"`
	Patch struct {
	} `command:"patch" description:"increase the patch version"`
	Init struct {
	} `command:"init" description:"init new version"`
	Date struct {
	} `command:"date" description:"print the current date and time in the format YYYYMMDDHHMMSS"`
	Auto struct {
		ExitCode bool `short:"e" long:"exit-code-if-no-bump" description:"exit code to use if no notable changes are found"`
	} `command:"auto" description:"check the git repository and increase the version if necessary. This implies --git"`
	Predict struct {
		ExitCode bool `short:"e" long:"exit-code-if-no-bump" description:"exit code to use if no notable changes are found"`
	} `command:"predict" description:"predict the next version based on the git history. This implies --git"`
	Print struct {
		ExitCode bool `short:"e" long:"exit-code-on-error" description:"exit code to use if an error occurs"`
	} `command:"print" description:"print the current version, you can combine this with --git to print the last tag"`
	Version struct {
	} `command:"version" description:"print the version of this tool"`
}

func increasePatch() (string, error) {
	version, err := getVersion()
	if err != nil {
		return "", err
	}

	next := version.IncPatch()
	return next.String(), nil
}

func increaseMinor() (string, error) {
	version, err := getVersion()
	if err != nil {
		return "", err
	}

	next := version.IncMinor()
	return next.String(), nil

}

func increaseMajor() (string, error) {
	version, err := getVersion()
	if err != nil {
		return "", err
	}

	next := version.IncMajor()
	return next.String(), nil
}

var verbosity = false
var nullTerminated = false
var command string
var predict = false

func executeCommand() {

	arguments = new(commandLineOptions)
	p := flags.NewParser(arguments, flags.Default)

	var err error

	_, err = p.Parse()
	if err != nil {
		if arguments.Verbose {
			fmt.Printf("Error: %s\n", err)
		}

		os.Exit(1)
	}

	activeCommand := p.Command.Active

	if activeCommand == nil {
		return
	}

	var newVersion string
	command = activeCommand.Name

	if arguments.Verbose {
		verbosity = true
	}

	if arguments.Null {
		nullTerminated = true
	}

	if command == "version" {
		fmt.Println("Name:     ", GetName())
		fmt.Println("Version:  ", GetVersion())
		fmt.Println("Build:    ", GetBuild())
		fmt.Println("Commit:   ", GetCommit())
		fmt.Println("Mnemonic: ", GetMnemonic())
		return
	}

	if command == "predict" {
		predict = true
	}

	if command == "auto" || command == "predict" {
		arguments.Git = true

		if arguments.Path != "" || arguments.Selector != "" {
			msg := "Error: cannot use --path or --selector with auto or predict\n"
			_, err := fmt.Fprintf(os.Stderr, msg)
			if err != nil {
				fmt.Printf(msg)
			}
			os.Exit(1)
		}

	}

	if arguments.Git {

		path, err := os.Getwd()
		if err != nil {
			_, err := fmt.Fprintf(os.Stderr, "Could not get current working directory: %s\n", err)
			if err != nil {
				fmt.Printf("Could not get current working directory: %s\n", err)
			}
			os.Exit(1)
		}

		if verbosity {
			fmt.Println("git path:", path)
		}

		gitRepo, err = git.PlainOpen(path)
		if err != nil {
			_, err := fmt.Fprintf(os.Stderr, "Error: %s\n", err)
			if err != nil {
				fmt.Printf("Error: %s\n", err)
			}
			os.Exit(1)
		}
	}

	if command == "auto" || command == "predict" {
		commitType, err := GetCommitType()

		if arguments.Verbose {
			fmt.Printf("commit type: %s\n", commitType)
		}

		if err != nil {
			_, err := fmt.Fprintf(os.Stderr, "Error: %s\n", err)
			if err != nil {
				fmt.Printf("Error: %s\n", err)
			}
			os.Exit(1)
		}

		switch commitType {
		case BreakingCommit:
			command = "major"
		case FeatCommit:
			command = "minor"
		case FixCommit:
			command = "patch"
		case OtherCommit:

			exitCode := 0
			printMessage := true
			if arguments.Auto.ExitCode || arguments.Predict.ExitCode {

				if !arguments.Verbose {
					printMessage = false
				}

				exitCode = 1
			}

			if printMessage {
				fmt.Println("No notable changes found.")
				fmt.Println("If you want to force a version\nincrease, use the major, minor\nor patch command.")
			}

			os.Exit(exitCode)
		}

	}

	switch command {
	case "print":

		version, err := getVersion()
		if err != nil {

			if arguments.Print.ExitCode {
				os.Exit(10)
			}

			_, err := fmt.Fprintf(os.Stderr, "Error: %s\n", err)
			if err != nil {
				fmt.Printf("Error: %s\n", err)
			}

			os.Exit(1)
		}

		if nullTerminated {
			// terminate with null byte
			fmt.Printf("%s%s", version.String(), string(rune(0)))
		} else {
			fmt.Printf("%s\n", version.String())
		}
		os.Exit(0)

	case "date":
		currentTime := time.Now()
		build = currentTime.Format("20060102150405")

		if nullTerminated {
			build += string(rune(0))
		}

		fmt.Printf("%s", build)
		os.Exit(0)

	case "init":
		newVersion = "0.1.0"
	case "major":
		newVersion, err = increaseMajor()
	case "minor":
		newVersion, err = increaseMinor()
	case "patch":
		newVersion, err = increasePatch()
	}

	if err != nil {
		_, err := fmt.Fprintf(os.Stderr, "Error: %s\n", err)
		if err != nil {
			fmt.Printf("Error: %s\n", err)
		}
		os.Exit(1)
	}

	if verbosity {
		fmt.Printf("new version: %s\n", newVersion)
	}

	err = writeVersion(newVersion)
	if err != nil {

		if activeCommand.Name == "auto" || activeCommand.Name == "predict" {
			if arguments.Auto.ExitCode || arguments.Predict.ExitCode {
				os.Exit(10)
			}
		}

		_, err := fmt.Fprintf(os.Stderr, "Error: %s\n", err)
		if err != nil {
			fmt.Printf("Error: %s\n", err)
		}
		os.Exit(1)
	}

	if verbosity {
		fmt.Printf("wrote version: %s\n", newVersion)
	}

}
